module CabezaDeTermo
	module Lunfardo
		class Scope
			class << self
				def new_scope_class(name:, &block)
					Class.new(self) do
						set_scope_name name
						infere_block_as_parameter_from(block)
					end
				end

				# Accessing

				def set_scope_name(name)
					@scope_name = name.to_sym
				end

				def scope_name()
					@scope_name
				end

				def infere_block_as_parameter_from(block)
					@block_as_parameter = 
						!block.nil? && !block.parameters.empty? &&
						block.parameters.last[0] == :block
				end

				def block_as_parameter?()
					@block_as_parameter
				end

				def scopes()
					@scopes ||= Hash[]
				end

				# Defining the DSL

				def define(method_name, &block)
					define_method(method_name, &block)
				end

				def before(*params, &block)
					define(:before, &block)
				end

				def perform(*params, &block)
					define(:perform, &block)
				end

				def on(method_name, &block)
					scopes[method_name] = Scope.new_scope_class(name: method_name, &block)

					scopes[method_name].instance_exec(&block)
				end

				def after(*params, &block)
					define(:after, &block)
				end

				def recursive(scope_name, scope:)
					scopes[scope_name] = scope
				end

				# Raising errors

				def raise_error(message)
					raise Lunfardo::Error.new(message)
				end
			end

			# Initializing

			def initialize(context, outer_scope)
				@context = context
				@outer_scope = outer_scope
			end

			# Accessing

			def _scope_at(scope_name)
				_scopes.fetch(scope_name.to_sym)
			end

			def _scope_name()
				self.class.scope_name
			end

			def _block_as_parameter?()
				self.class.block_as_parameter?
			end

			def _scopes()
				self.class.scopes
			end

			# Asking

			def _includes_scope?(scope_name)
				_scopes.key?(scope_name.to_sym)
			end

			# Evaluating

			def context()
				@context
			end

			def outer(n = 1)
				(1..n-1).inject(@outer_scope) do |outer_scope, i|
					outer_scope.outer
				end
			end

			def _evaluate(*params, &block)
				result = send(:before, *params, &block) if respond_to?(:before)

				result = send(:perform, *params, &block) if respond_to?(:perform)

				result = instance_exec(*params, &block) unless block.nil? || _block_as_parameter?

				result = send(:after, *params, &block) if respond_to?(:after)

				result
			end

			def method_missing(method_name, *params, &block)
				return _raise_missing_scope_error(method_name) unless _includes_scope?(method_name)

				_scope_at(method_name).new(@context, self)._evaluate(*params, &block)
			end

			# Raising errors

			def _raise_missing_scope_error(method_name)
				path = (_scope_path << method_name).join('.')
				self.class.raise_error("The scope '#{path}' is not defined")
			end

			def _scope_path
				outer.nil? ? [_scope_name] : outer._scope_path << _scope_name
			end
		end
	end
end